use std::collections::HashMap;

const ASXVERSION: [u8; 3] = [0,1,0];

#[derive(Debug, Clone)]
pub enum VMMemory {
    String(VMMemoryString),
    Number(VMMemoryNumber),
    Boolean(VMMemoryBoolean),
    Function(VMMemoryFunction),
    Null(VMMemoryNull),
}

#[derive(Debug, Clone)]
struct VMMemoryString {
    value: String,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryNumber {
    value: i64,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryBoolean {
    value: bool,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryFunction {
    function: DecompiledFunction,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryNull {
    variable_id: u32,
}

#[derive(Debug, Clone)]
struct DecompiledFile {
    name: String,
    functions: Vec<DecompiledFunction>,
}
#[derive(Debug, Clone)]
struct DecompiledFunction {
    name: String,
    body: Vec<DecompiledOperation>,
    var_ids: HashMap<String, u32>,
    strings: HashMap<u32, String>,
    functions: HashMap<u32, u32>
}
#[derive(Debug, Clone)]
struct DecompiledOperation {
    opcode: u8,
    arg1: u8,
    arg2: i64,
    arg3: u8,
}

#[derive(Debug, Clone)]
pub struct Register {
    id: u8,
    pointer: usize
}

pub struct Machine {
    pub memory: Vec<VMMemory>,
    pub files: Vec<DecompiledFile>,
    pub stack: Vec<VMMemory>,
    pub registers: Vec<Register>,
    pub pc: usize,
}

fn read_be_num(input: &[u8]) -> usize {
    let mut result: usize = 0;
    for i in 0..input.len() {
        result <<= 8;
        result |= input[i] as usize;
    }
    return result;
}

fn read_str(input: &[u8]) -> String {
    let mut result = String::new();
    for &byte in input {
        result.push(byte as char);
    }
    return result;
}

fn read_bin(input: &[u8]) -> String {
    let mut result = String::new();
    for &byte in input {
        result.push_str(&format!("{:08b}", byte));
    }
    return result;
}

fn bin_to_num(bin: String) -> usize {
    let mut result: usize = 0;
    for (i, c) in bin.chars().rev().enumerate() {
        if c == '1' {
            result += 1 << i;
        }
    }
    return result;
}

fn bin_to_snum(bin: String) -> i64 {
    let mut result: i64 = 0;
    let mut is_negative = false;
    for (i, c) in bin.chars().rev().enumerate() {
        if c == '1' {
            if i == bin.len() - 1 {
                is_negative = true;
            } else {
                result += 1 << i;
            }
        }
    }
    if is_negative {
        result = -result;
    }
    return result;
}

fn get_file_by_name(files: &Vec<DecompiledFile>, name: &str) -> Option<DecompiledFile> {
    for file in files {
        if file.name == name {
            return Some(file.clone());
        }
    }
    None
}

fn get_function_by_name(file: &DecompiledFile, name: &str) -> Option<DecompiledFunction> {
    for function in &file.functions {
        if function.name == name {
            return Some(function.clone());
        }
    }
    None
}

fn get_register_by_id(registers: &Vec<Register>, id: u8) -> Option<&Register> {
    for reg in registers {
        if reg.id == id {
            return Some(reg);
        }
    }
    None
}

fn set_register(registers: &mut Vec<Register>, new_register: Register) {
    if new_register.id == 0 {
        return;
    }
    for (i, reg) in registers.iter().enumerate() {
        if reg.id == new_register.id {
            registers[i] = new_register;
            return;
        }
    }
}

fn get_mem_pos_by_var_id(memory: &Vec<VMMemory>, var_id: u32) -> Option<usize> {
    for i in 0..memory.len() {
        match &memory[i] {
            VMMemory::String(str) if str.variable_id == var_id => return Some(i),
            VMMemory::Number(num) if num.variable_id == var_id => return Some(i),
            VMMemory::Boolean(bool) if bool.variable_id == var_id => return Some(i),
            VMMemory::Null(null) if null.variable_id == var_id => return Some(i),
            VMMemory::Function(func) if func.variable_id == var_id => return Some(i),
            _ => continue
        }
    }
    None
}

fn do_operation_operation(registers: &mut Vec<Register>, memory: &mut Vec<VMMemory>, operation: &DecompiledOperation) {
    let reg1 = get_register_by_id(registers, operation.arg1);
    let reg2 = get_register_by_id(registers, operation.arg2 as u8);
    let reg3 = get_register_by_id(registers, operation.arg3);
    if reg1.is_none() || reg2.is_none() || reg3.is_none() {
        panic!("One or more registers not found for operation");
    }
    let mem1 = reg1.unwrap().pointer;
    let mem2 = reg2.unwrap().pointer;
    let mut mem3 = reg3.unwrap().pointer;
    if memory.get(mem1).is_none() || memory.get(mem2).is_none() {
        panic!("Memory location not found for registers");
    }
    if reg3.unwrap().pointer == 0 {
        match operation.opcode {
            10 => {
                match memory.get(mem1).unwrap() {
                    VMMemory::String(_) => {
                        memory.push(VMMemory::String(VMMemoryString { value: String::new(), variable_id: 0 }));
                        set_register(registers, Register { id: operation.arg3, pointer: memory.len() - 1 });
                        mem3 = memory.len() - 1;
                    }
                    VMMemory::Number(_) => {
                        memory.push(VMMemory::Number(VMMemoryNumber { value: 0, variable_id: 0 }));
                        set_register(registers, Register { id: operation.arg3, pointer: memory.len() - 1 });
                        mem3 = memory.len() - 1;
                    }
                    _ => {
                        panic!("Invalid memory type for operation");
                    }
                }
            },
            11 | 12 | 13 | 14 | 15 => {
                memory.push(VMMemory::Number(VMMemoryNumber { value: 0, variable_id: 0 }));
                set_register(registers, Register { id: operation.arg3, pointer: memory.len() - 1 });
                mem3 = memory.len() - 1;
            },
            16 | 17 | 18 | 19 | 20 | 21 | 22 | 23 => {
                memory.push(VMMemory::Boolean(VMMemoryBoolean { value: false, variable_id: 0 }));
                set_register(registers, Register { id: operation.arg3, pointer: memory.len() - 1 });
                mem3 = memory.len() - 1;
            },
            _ => {}
        }
    }
    let temp_mem = memory.clone();
    match (&temp_mem[mem1], &temp_mem[mem2], &mut memory[mem3]) {
        (VMMemory::Number(num1), VMMemory::Number(num2), VMMemory::Number(num3)) => {
            match operation.opcode {
                10 => {
                    num3.value = num1.value + num2.value;
                },
                11 => {
                    num3.value = num1.value - num2.value;
                },
                12 => {
                    num3.value = num1.value * num2.value;
                },
                13 => {
                    if num2.value == 0 {
                        panic!("Division by zero");
                    }
                    num3.value = num1.value / num2.value;
                },
                14 => {
                    num3.value = num1.value.pow(num2.value as u32);
                },
                15 => {
                    num3.value = num1.value % num2.value;
                },
                _ => {
                    panic!("Invalid operation for number operation")
                }
            }
        },
        (VMMemory::Number(num1), VMMemory::Number(num2), VMMemory::Boolean(bool3)) => {
            match operation.opcode {
                18 => {
                    bool3.value = num1.value == num2.value;
                },
                19 => {
                    bool3.value = num1.value != num2.value;
                },
                20 => {
                    bool3.value = num1.value > num2.value;
                },
                21 => {
                    bool3.value = num1.value >= num2.value;
                },
                22 => {
                    bool3.value = num1.value < num2.value;
                },
                23 => {
                    bool3.value = num1.value <= num2.value;
                },
                _ => panic!("Invalid operation for boolean operation"),
            }
        },
        (VMMemory::String(str1), VMMemory::String(str2), VMMemory::Boolean(bool3)) => {
            match operation.opcode {
                18 => {
                    bool3.value = str1.value == str2.value;
                },
                19 => {
                    bool3.value = str1.value != str2.value;
                },
                _ => panic!("Invalid operation for string operation"),
            }
        },
        (VMMemory::String(str1), VMMemory::String(str2), VMMemory::String(str3)) => {
            match operation.opcode {
                10 => {
                    str3.value = str1.value.clone() + &str2.value;
                },
                _ => panic!("Invalid operation for string operation"),
            }
        },
        (VMMemory::Boolean(bool1), VMMemory::Boolean(bool2), VMMemory::Boolean(bool3)) => {
            match operation.opcode {
                16 => {
                    bool3.value = bool1.value && bool2.value;
                },
                17 => {
                    bool3.value = bool1.value || bool2.value;
                },
                _ => panic!("Invalid operation for boolean operation"),
            }
        },
        _ => panic!("Invalid memory types for operation"),
    }
}

impl Machine {
    pub fn new() -> Self {
        let mut registers = Vec::new();
        for i in 0..17 {
            registers.push(Register {
                id: i,
                pointer: 0,
            });
        }
        return Machine {
            memory: vec![
                VMMemory::Null(VMMemoryNull { variable_id: 0 })
            ],
            files: Vec::new(),
            stack: Vec::new(),
            registers: registers,
            pc: 0,
        };
    }

    pub fn load(&mut self, data: Vec<u8>) {
        if data[0..3] != *"ASX".as_bytes() {
            panic!("Invalid ASX file header");
        }
        if data[3..6] != ASXVERSION {
            panic!("Unsupported ASX version");
        }
        let filename_len = read_be_num(&data[6..10]);
        let filename = read_str(&data[10..10 + filename_len]);
        let mut offset = 10 + filename_len;
        let func_name_len = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        let func_name = read_str(&data[offset..offset + func_name_len]);
        offset += func_name_len;
        let var_id_len = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        let mut var_ids: HashMap<String, u32> = HashMap::new();
        for _ in 0..var_id_len {
            let name_len = read_be_num(&data[offset..offset + 4]);
            offset += 4;
            let name = read_str(&data[offset..offset + name_len]);
            offset += name_len;
            let var_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            var_ids.insert(name, var_id);
        }
        let mut strings: HashMap<u32, String> = HashMap::new();
        let string_count = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0..string_count {
            let str_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            let str_len = read_be_num(&data[offset..offset + 4]);
            offset += 4;
            let str_value = read_str(&data[offset..offset + str_len]);
            offset += str_len;
            strings.insert(str_id, str_value);
        }
        let mut functions: HashMap<u32, u32> = HashMap::new();
        let _function_count = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0.._function_count {
            let func_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            let func_pos = read_be_num(&data[offset..offset + 4]) as u32;
            offset += 4;
            functions.insert(func_id, func_pos);
        }
        let mut body: Vec<DecompiledOperation> = Vec::new();
        let instr_len: usize = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0..instr_len {
            let op = read_bin(&data[offset..offset + 10]);
            offset += 10;
            println!("{}", op);
            let op_code = bin_to_num(op[0..5].to_string());
            let arg1 = bin_to_num(op[5..9].to_string());
            let arg2 = bin_to_snum(op[9..73].to_string());
            let arg3 = bin_to_num(op[73..77].to_string());
            body.push(DecompiledOperation {
                opcode: op_code as u8,
                arg1: arg1 as u8,
                arg2: arg2,
                arg3: arg3 as u8,
            });
        }
        let mut file_functions: Vec<DecompiledFunction> = Vec::new();
        file_functions.push(DecompiledFunction { name: func_name, body: body, var_ids: var_ids, strings: strings, functions: functions });
        self.files.push(DecompiledFile {
            name: filename,
            functions: file_functions
        });
        println!("Decompiled: {:?}", self.files[0]);
    }

    pub fn run(&mut self) {
        let main = get_file_by_name(&self.files, "main.asl");
        if main.is_none() {
            panic!("Main file not found");
        }
        let main = main.unwrap();
        let main_func = get_function_by_name(&main, "main");
        if main_func.is_none() {
            panic!("Main function not found");
        }
        let main_func = main_func.unwrap();
        while self.pc < main_func.body.len() {
            let operation = &main_func.body[self.pc];
            self.pc += 1;
            match operation.opcode {
                0 => {
                    //HLT
                    return;
                },
                1 => {
                    //LDS
                    let str = main_func.strings.get(&(operation.arg2 as u32));
                    if str.is_none() {
                        panic!("String with ID {} not found", operation.arg2);
                    }
                    let str = str.unwrap();
                    self.memory.push(VMMemory::String(VMMemoryString { value: str.clone(), variable_id: 0 }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                2 => {
                    //LDM
                    let mem = get_mem_pos_by_var_id(&self.memory, operation.arg2 as u32);
                    if mem.is_none() {
                        panic!("Memory with variable ID {} not found", operation.arg2);
                    }
                    let mem = mem.unwrap();
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: mem });
                },
                3 => {
                    //LDI
                    self.memory.push(VMMemory::Number(VMMemoryNumber { value: operation.arg2, variable_id: 0 }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                4 => {
                    //LDB
                    self.memory.push(VMMemory::Boolean(VMMemoryBoolean { value: operation.arg2 != 0, variable_id: 0 }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                5 => {
                    //LDF
                    let func_pos = main_func.functions.get(&(operation.arg2 as u32));
                    if func_pos.is_none() {
                        panic!("Function with ID {} not found", operation.arg2);
                    }
                    let func_pos = func_pos.unwrap();
                    let func = main.functions.get(*func_pos as usize);
                    if func.is_none() {
                        panic!("Function with position {} not found", func_pos);
                    }
                    self.memory.push(VMMemory::Function(VMMemoryFunction {
                        function: func.unwrap().clone(),
                        variable_id: 0,
                    }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                6 => {
                    //LDN
                    self.memory.push(VMMemory::Null(VMMemoryNull { variable_id: 0 }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                7 => {
                    //ASS
                    let reg = get_register_by_id(&self.registers, operation.arg1);
                    if reg.is_none() {
                        panic!("Register {} not found", operation.arg1);
                    }
                    let reg = reg.unwrap();
                    if reg.pointer >= self.memory.len() || reg.pointer < 1 {
                        panic!("Register {} points to an invalid memory location", operation.arg1);
                    }
                    match &mut self.memory[reg.pointer] {
                        VMMemory::Number(num) => {
                            num.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::String(str) => {
                            str.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::Boolean(bool) => {
                            bool.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::Null(null) => {
                            null.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::Function(func) => {
                            func.variable_id = operation.arg2 as u32;
                        },
                    }
                },
                8 => {
                    //UNB
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: 0 });
                },
                9 => {
                    //MOV
                    let temp_regs = self.registers.clone();
                    let reg1 = get_register_by_id(&temp_regs, operation.arg1);
                    if reg1.is_none() {
                        panic!("Register {} not found", operation.arg1);
                    }
                    let reg1 = reg1.unwrap();
                    set_register(&mut self.registers, Register { id: operation.arg2 as u8, pointer: reg1.pointer });
                },
                10 | 11 | 12 | 13 | 14 | 15 | 16 | 17 | 18 | 19 | 20 | 21 | 22 | 23 => {
                    do_operation_operation(&mut self.registers, &mut self.memory, operation);
                },
                24 => {
                    //NOT
                    let reg1 = get_register_by_id(&self.registers, operation.arg1);
                    let reg2 = get_register_by_id(&self.registers, operation.arg2 as u8);
                    if reg1.is_none() || reg2.is_none() {
                        panic!("One or more registers not found for NOT operation");
                    }
                    let mem1 = reg1.unwrap().pointer;
                    let mut mem2 = reg2.unwrap().pointer;
                    if self.memory.get(mem1).is_none() {
                        panic!("Memory location not found for register {}", operation.arg1);
                    }
                    if reg2.unwrap().pointer == 0 {
                        self.memory.push(VMMemory::Boolean(VMMemoryBoolean { value: false, variable_id: 0 }));
                        set_register(&mut self.registers, Register { id: operation.arg3, pointer: self.memory.len() - 1 });
                        mem2 = self.memory.len() - 1;
                    }
                    let temp_mem = self.memory.clone();
                    match (&temp_mem[mem1], &mut self.memory[mem2]) {
                        (VMMemory::Boolean(bool1), VMMemory::Boolean(bool2)) => {
                            bool2.value = !bool1.value;
                        },
                        _ => panic!("Invalid memory types for NOT operation"),
                    }
                },
                25 => {
                    //JMP
                    self.pc = operation.arg2 as usize;
                },
                26 => {
                    //CJP
                    let reg = get_register_by_id(&self.registers, operation.arg1);
                    if reg.is_none() {
                        panic!("Register {} not found", operation.arg1);
                    }
                    let reg = reg.unwrap();
                    let mem = self.memory.get(reg.pointer);
                    if mem.is_none() {
                        panic!("Memory location not found for register");
                    }
                    let mem = mem.unwrap();
                    match &mem {
                        VMMemory::Boolean(bool) => {
                            if bool.value {
                                self.pc = operation.arg2 as usize;
                            }
                        }
                        _ => panic!("Invalid memory type for CJP operation"),
                    }
                },
                27 => {
                    //CAL
                    //not implemented yet
                },
                28 => {
                    //PSH
                    let reg = get_register_by_id(&self.registers, operation.arg1);
                    if reg.is_none() {
                        panic!("Register {} not found", operation.arg1);
                    }
                    let reg = reg.unwrap();
                    let mem = self.memory.get(reg.pointer);
                    if mem.is_none() {
                        panic!("Memory location not found for register");
                    }
                    let mem = mem.unwrap();
                    self.stack.push(mem.clone());
                },
                29 => {
                    //RET
                    //not implemented yet
                },
                30 => {
                    //not implemented yet
                },
                31 => {
                    //GET
                    //not implemented yet
                },
                32 => {
                    //SET
                    //not implemented yet
                },
                _ => {
                    panic!("Unknown operation code: {}", operation.opcode);
                }
            }
        }
    }
}